import { addGlobalEventProcessor, getCurrentHub } from "../../hub/src/index.ts";
import { __DEBUG_BUILD__ } from "../../types/src/globals.ts";
import { Integration, Options } from "../../types/src/index.ts";
import { logger } from "../../utils/src/index.ts";

export const installedIntegrations: string[] = [];

/** Map of integrations assigned to a client */
export type IntegrationIndex = {
  [key: string]: Integration;
};

/**
 * @private
 */
function filterDuplicates(integrations: Integration[]): Integration[] {
  return integrations.reduce((acc, integrations) => {
    if (
      acc.every((accIntegration: any) =>
        integrations.name !== accIntegration.name
      )
    ) {
      acc.push(integrations);
    }
    return acc;
  }, [] as Integration[]);
}

/** Gets integration to install */
export function getIntegrationsToSetup(options: Options): Integration[] {
  const defaultIntegrations =
    (options.defaultIntegrations && [...options.defaultIntegrations]) || [];
  const userIntegrations = options.integrations;

  let integrations: Integration[] = [...filterDuplicates(defaultIntegrations)];

  if (Array.isArray(userIntegrations)) {
    // Filter out integrations that are also included in user options
    integrations = [
      ...integrations.filter((integrations) =>
        userIntegrations.every((userIntegration) =>
          userIntegration.name !== integrations.name
        )
      ),
      // And filter out duplicated user options integrations
      ...filterDuplicates(userIntegrations),
    ];
  } else if (typeof userIntegrations === "function") {
    integrations = userIntegrations(integrations);
    integrations = Array.isArray(integrations) ? integrations : [integrations];
  }

  // Make sure that if present, `Debug` integration will always run last
  const integrationsNames = integrations.map((i) => i.name);
  const alwaysLastToRun = "Debug";
  if (integrationsNames.indexOf(alwaysLastToRun) !== -1) {
    integrations.push(
      ...integrations.splice(integrationsNames.indexOf(alwaysLastToRun), 1),
    );
  }

  return integrations;
}

/**
 * Given a list of integration instances this installs them all. When `withDefaults` is set to `true` then all default
 * integrations are added unless they were already provided before.
 * @param integrations array of integration instances
 * @param withDefault should enable default integrations
 */
export function setupIntegrations(
  integrations: Integration[],
): IntegrationIndex {
  const integrationIndex: IntegrationIndex = {};

  integrations.forEach((integration) => {
    integrationIndex[integration.name] = integration;

    if (installedIntegrations.indexOf(integration.name) === -1) {
      integration.setupOnce(addGlobalEventProcessor, getCurrentHub);
      installedIntegrations.push(integration.name);
      __DEBUG_BUILD__ &&
        logger.log(`Integration installed: ${integration.name}`);
    }
  });

  return integrationIndex;
}
